import re


def is_number(s):
    """
    判断字符串是否是数字
    :param s: 字符串
    :return: True/False
    """
    if s.count('.') < 2 and s.replace('.', '').isdigit():
        return True
    return False


def is_variable_name(s):
    """
    判断字符串是否是一个变量名
    :param s: 字符串
    :return: True/False
    """
    if s is None:
        return False
    m = re.match(r"^\d.*", s)
    if m is not None:
        return False
    wrong = [')', '(', '+', '-', '*', '/', '?', '!', ':', '>', '<', '=', '^']
    for w in wrong:
        if s.find(w) != -1:
            return False
    return True
    # m = re.match("^[_a-zA-Z\w]*$", s)
    # return False if m is None else True




import math

from lexp.operation import Operation

# 二元运算符


class Plus(Operation):
    # 加法运算
    name = '+'
    argc = 2

    def operate(self, argv):
        super().check(argv)
        return float(argv[0]) + float(argv[1])


class Minus(Operation):
    # 减法运算
    name = '-'
    argc = 2

    def operate(self, argv):
        super().check(argv)
        return float(argv[0]) - float(argv[1])


class Multi(Operation):
    # 乘法运算
    name = '*'
    argc = 2

    def operate(self, argv):
        super().check(argv)
        return float(argv[0]) * float(argv[1])


class Divide(Operation):
    # 除法运算
    name = '/'
    argc = 2

    def operate(self, argv):
        super().check(argv)
        f1 = float(argv[1])
        if f1 == 0:
            return None
        return float(argv[0]) / float(argv[1])


class Power(Operation):
    # 幂运算
    name = '^'
    argc = 2

    def operate(self, argv):
        super().check(argv)
        return math.pow(float(argv[0]), float(argv[1]))



import math

import numpy


class Functions:
    custom_function = dict()

    @classmethod
    def is_defined(cls, name):
        """
        该函数名是否已预先定义
        :param name: 函数名称
        :return: bool
        """
        if not hasattr(cls, name):
            return False
        func = getattr(cls, name)
        return True if callable(func) else False

    @classmethod
    def get_defined(cls, name):
        func = getattr(cls, name, None)
        if not callable(func):
            raise Exception(f"{name} not callable")
        return func

    # 以下为lexp预先定义的函数

    @staticmethod
    def pow(x, a):
        return x ** a
        # return math.pow(x, a)

    @staticmethod
    def sin(x):
        return math.sin(x)

    @staticmethod
    def cos(x):
        return math.cos(x)

    @staticmethod
    def log(x, a):
        return math.log(x, a)

    @classmethod
    def avg(cls, *nums):
        n = len(nums)
        return cls.sum(*nums) / n

    @staticmethod
    def sum(*nums):
        total = 0
        for num in nums:
            total = total + float(num)
        return total

    @staticmethod
    def min(*vals):
        min_val = None
        for val in vals:
            val = float(val)
            if min_val is None or val < min_val:
                min_val = val
        return min_val

    @staticmethod
    def max(*vals):
        max_val = None
        for val in vals:
            val = float(val)
            if max_val is None or val > max_val:
                max_val = val
        return max_val

    @staticmethod
    def abs(val):
        return abs(val)

    @staticmethod
    def sqrt(val):
        return math.sqrt(val)

    @staticmethod
    def cbrt(val):
        return numpy.cbrt(val)

    @staticmethod
    def ifnull(val, default_val):
        """
        val为None时返回一个默认值
        :param val: 数值
        :param default_val: 默认值
        :return: 检查后的返回值
        """
        if val is None:
            return default_val
        return val

    @staticmethod
    def nullif(val, target_val):
        """
        如果val等于某个特定值，返回None
        :param val: 待检查的值
        :param target_val: 特定值
        :return: 结果
        """
        if val == target_val:
            return None
        return val

    @staticmethod
    def coalesce(*vals):
        """
        返回所有参数中第一个非空非零的值
        :param vals: 所有参数
        :return: float
        """
        for val in vals:
            if val is None:
                continue
            val = float(val)
            if val > 0:
                return val
        return vals[0]



import re

from lexp.const import defined_constant
from lexp.exception import ParseError, CalcError
from lexp.funcs import Functions
from lexp.operation.binary import Plus, Minus, Multi, Divide, Power
from lexp.operation.bool import Or, And, Eq, Neq, Ge, Le, Gt, Lt
from lexp.operation.custom import CustomOperation, Variable
from lexp.operation.ternary import EitherOr
from lexp.operation.unary import Not, Negative
from lexp.util import is_number, is_variable_name


class _Part:
    """
    表达式被切分后的组成部分，有上述定义的类型
    """

    def __init__(self, string, t, name=''):
        self.string = string
        self.type = t
        self.name = name

    def __str__(self):
        return self.string


class _Node:
    # 以下为Part类型
    # 函数
    FUNCTION = 1
    # 变量
    VARIABLE = 2
    # 运算符
    OPERATOR = 3
    # 常量
    CONSTANT = 4
    # 括号
    BRACKETS = 5
    # 取负
    NEGATIVE = 6

    def __init__(self, string=''):
        # operation为None，则未解析
        # 此为该节点的运算方式
        self.operation = None

        # 该节点的原始字符串
        self.string = re.sub('[ \r\n\t]', '', string)

        # 参数节点列表，当使用operation进行运算时，将会把所有children带入
        self.children = list()

        # value不为None，则为常量
        self.value = None

    def get_functions(self):
        """
        获得节点中所有函数的信息
        :return: list
        """
        arr = []
        for i in range(0, len(self.children)):
            arr.extend(self.children[i].get_functions())
        if isinstance(self.operation, CustomOperation):
            name = self.operation.name
            if Functions.is_defined(name):
                # lexp预先定义
                t = 'preset'
            else:
                # 用户自行定义
                t = 'custom'
            arr.append({
                'type': t,
                'name': name,
                'minArgc': self.operation.argc
            })
        return arr

    def get_variables(self):
        """
        获得表达式中出现的所有变量
        :return: list
        """
        arr = []
        for i in range(0, len(self.children)):
            child_arr = self.children[i].get_variables()
            for v in child_arr:
                if v not in arr:
                    arr.append(v)
        if isinstance(self.operation, Variable):
            name = self.operation.variable
            if name not in arr:
                arr.append(name)
        return arr

    @staticmethod
    def _get_bracket_position(s):
        """
        最外层括号的位置
        :param s: 字符串
        :return: list
        """
        positions = []
        stack = []
        it = re.finditer('[(]|[)]', s)
        arr = [m.span()[0] for m in it]
        for i in range(0, len(arr)):
            p = arr[i]
            c = s[p:p + 1]
            if c == '(':
                stack.append(p)
            elif c == ')':
                p0 = stack.pop()
                if len(stack) == 0:
                    positions.append((p0, p))
        return positions

    def get_argv(self, string):
        """
        获取字符串参数列表
        :param string: 字符串
        :return: list
        """
        if string == '':
            return []
        # 以最外层括号分块，保持括号中的内容为一个整体
        outer_brackets = self._get_bracket_position(string)
        parts = []
        p = 0
        for i in range(0, len(outer_brackets)):
            bracket_position = outer_brackets[i]
            b0 = bracket_position[0]
            b1 = bracket_position[1]
            if b0 > p:
                parts.append(_Part(string[p:b0], 0))
            parts.append(_Part(string[b0 + 1: b1], self.BRACKETS))
            p = b1 + 1
        if len(string) > p:
            parts.append(_Part(string[p:], 0))
        # 获取以逗号分隔的参数列表，除括号以外
        # argv为参数列表，每一项为part的数组
        argv = []
        # last用作临时存储之前的可能会和后面作为一体的part，比如func与(x+y)
        last = []
        for i in range(0, len(parts)):
            part = parts[i]
            s = part.string
            if part.type == self.BRACKETS:
                last.append(part)
                continue
            ps = [_Part(x, 0) for x in s.split(',')]
            if len(ps) == 1:
                last.append(part)
                continue
            size = len(ps)
            for j in range(0, size):
                p = ps[j]
                if j == 0:
                    if p.string != '':
                        last.append(p)
                    argv.append(last)
                elif j == size - 1:
                    if p.string == '':
                        last = []
                    else:
                        last = [p]
                else:
                    if p.string != '':
                        argv.append([p])
        argv.append(last)
        return argv

    def parse_operation(self, parts):
        """
        解析其中一项参数的字符数组
        :param parts: part list
        :return: None
        """
        characters = []
        for i in range(0, len(parts)):
            p = parts[i]
            s = p.string
            if p.type == self.BRACKETS:
                characters.append(p)
                continue
            it = re.finditer('[-+/*!?:&|=><^]', s)
            position = 0
            for m in it:
                pos = m.span()
                p0 = pos[0]
                p1 = pos[1]
                if p0 > position:
                    characters.append(_Part(s[position:p0], 0))
                characters.append(_Part(s[p0:p1], self.OPERATOR))
                position = p1
            if len(s) > position:
                characters.append(_Part(s[position:len(s)], 0))

        last = None
        arr = []

        def add_single_part(_p):
            if _p is None:
                return
            _c = _p.string
            if _p.type == _Node.BRACKETS:
                arr.append(_Part(_c, _Node.BRACKETS))
            elif is_number(_c):
                arr.append(_Part(_c, _Node.CONSTANT))
            elif is_variable_name(_c):
                arr.append(_Part(_c, _Node.VARIABLE))
            elif _c in ['+', '-', '*', '/', '?', '!', ':', '>', '<', '=', '^']:
                arr.append(_Part(_c, _Node.OPERATOR))
            else:
                # arr.append(_Part(_c, _Node.VARIABLE))
                raise ParseError('wrong charactors %s' % _c)

        # 合并运算符及函数，并给出类型
        for i in range(0, len(characters)):
            p = characters[i]
            c = p.string
            if last is None:
                last_c = None
                last_type = None
            else:
                last_c = last.string
                last_type = last.type

            if p.type == self.BRACKETS and last and is_variable_name(last.string):
                # 前为变量名，后为括号内容，这是一个函数
                arr.append(_Part(c, self.FUNCTION, last_c))
                last = None
            elif last_c == '|' and c == '|':
                arr.append(_Part('||', self.OPERATOR))
                last = None
            elif last_c == '=' and c == '=':
                arr.append(_Part('==', self.OPERATOR))
                last = None
            elif last_c == '!' and c == '=':
                arr.append(_Part('!=', self.OPERATOR))
                last = None
            elif last_c == '&' and c == '&':
                arr.append(_Part('&&', self.OPERATOR))
                last = None
            elif last_c == '>' and c == '=':
                arr.append(_Part('>=', self.OPERATOR))
                last = None
            elif last_c == '<' and c == '=':
                arr.append(_Part('<=', self.OPERATOR))
                last = None
            else:
                if last_type == self.OPERATOR and p.type == self.OPERATOR:
                    allow_repeat_operators = ['!', '-']
                    if c not in allow_repeat_operators and last_c not in allow_repeat_operators:
                        raise ParseError('operator error: %s' % c)
                add_single_part(last)
                last = p
        add_single_part(last)

        # 将符合条件的"-"标为负号
        for i in range(0, len(arr)):
            p = arr[i]
            if p.string == '-':
                if i == 0:
                    p.type = self.NEGATIVE
                elif arr[i - 1].type == self.OPERATOR:
                    p.type = self.NEGATIVE

        # 进一步将组合起来的Part解析为operation和children
        self.parse_single_part(arr)

    def parse_single_part(self, parts):
        """
        递归解析标识好类型的字符数组
        根据parts解析operation和children,形成一个功能完善的Node
        :param parts: part list
        :return: None
        """
        if len(parts) == 1:
            # 只有单个part的情况（常数、变量、括号内容、函数内容）
            p = parts[0]
            if p.type == self.CONSTANT:
                self.value = p.string
            elif p.type == self.VARIABLE:
                self.operation = Variable(p.string)
                self.string = p.string
            # 解析括号
            elif p.type == self.BRACKETS:
                self.string = p.string
                self.parse()
            # 解析函数
            elif p.type == self.FUNCTION:
                self.string = p.string
                self.parse(p.name)
            else:
                raise ParseError('unknown charactor %s' % p.string)
            return
        # 多个part组合的情况
        # 先解析三元运算符
        if self.split_by_ternary(parts):
            return
        # 再解析二元运算符
        if self.split_by_binary(parts):
            return
        # 再解析一元运算符
        if self.split_by_unary(parts):
            return
        # 以上都不是，那么出错了
        arr = [p.string for p in parts]
        raise ParseError('syntax error %s' % ' '.join(arr))

    def split_by_ternary(self, parts):
        """
        解析三元运算符
        只有一种情况: a ? b : c, 三个参数
        :param parts: part list
        :return: True/False
        """
        stack = []
        # 如有多个三元运算符嵌套，取最外层
        for i in range(0, len(parts)):
            p = parts[i]
            if p.type != self.OPERATOR:
                continue
            if p.string == '?':
                stack.append(i)
            elif p.string == ':':
                i0 = stack.pop()
                self.operation = EitherOr()
                condition = _Node()
                condition.parse_single_part(parts[0: i0])

                left = _Node()
                left.parse_single_part(parts[i0 + 1: i])

                right = _Node()
                right.parse_single_part(parts[i + 1:])
                self.children = [condition, left, right]
                return True
        if len(stack) != 0:
            raise ParseError('wrong tertiary operator')
        return False

    def split_by_binary(self, parts):
        """
        解析二元运算符
        :param parts: part list
        :return: True/False
        """
        # 二元运算符优先级(倒序)
        priority = ['||', '&&', '==', '!=', '>=', '<=', '>', '<', '+', '-', '*', '/', '^']
        cls = self.__class__
        for i in range(0, len(priority)):
            symbol = priority[i]
            for j in range(0, len(parts)):
                p = parts[j]
                if p.type != cls.OPERATOR or p.string != symbol:
                    continue
                if symbol == '||':
                    self.operation = Or()
                elif symbol == '&&':
                    self.operation = And()
                elif symbol == '==':
                    self.operation = Eq()
                elif symbol == '!=':
                    self.operation = Neq()
                elif symbol == '>=':
                    self.operation = Ge()
                elif symbol == '<=':
                    self.operation = Le()
                elif symbol == '>':
                    self.operation = Gt()
                elif symbol == '<':
                    self.operation = Lt()
                elif symbol == '+':
                    self.operation = Plus()
                elif symbol == '-':
                    self.operation = Minus()
                elif symbol == '*':
                    self.operation = Multi()
                elif symbol == '/':
                    self.operation = Divide()
                elif symbol == '^':
                    self.operation = Power()
                else:
                    return False
                # 所有二元运算符，一个在左一个在右，两个参数
                left = cls()
                left.parse_single_part(parts[0: j])
                right = cls()
                right.parse_single_part(parts[j + 1:])
                self.children = [left, right]
                return True
        return False

    def split_by_unary(self, parts):
        """
        解析一元运算符
        一元运算符只有两种情况(非、负号)
        :param parts: part list
        :return: Boolean
        """
        n_parts = len(parts)
        for i in range(0, n_parts):
            p = parts[i]
            if p.type == self.OPERATOR and p.string == '!':
                if i >= n_parts - 1:
                    raise ParseError('operator error:!')
                next_p = parts[i + 1]
                self.operation = Not()
                node = _Node()
                node.parse_single_part([next_p])
                self.children.append(node)
                return True
            elif p.type == self.NEGATIVE:
                if i >= n_parts - 1:
                    raise ParseError('operator error:-')
                next_p = parts[i + 1]
                self.operation = Negative()
                node = _Node()
                node.parse_single_part([next_p])
                self.children.append(node)
                return True
        return False

    def parse_children(self, argv):
        """
        解析函数中多项参数
        :param argv: 参数list,每一项为part list
        :return: None
        """
        self.children = []
        for i in range(0, len(argv)):
            parts = argv[i]
            node = _Node()
            node.parse_operation(parts)
            self.children.append(node)

    def parse(self, name=None):
        """
        解析括号或函数中的内容
        :param name: 括号区域的名称，如果没有则为普通括号，有则是函数名
        :return: self
        """
        if not isinstance(self.string, str):
            raise ParseError('no string to parse')
        argv = self.get_argv(self.string)
        if name is None:
            # 普通括号中，不应出现逗号分隔
            if len(argv) > 1:
                raise ParseError('syntax error: %s' % self.string)
            self.parse_operation(argv[0])
        else:
            # 函数中的内容
            if name in Functions.custom_function:
                # 已经注册，归为自定义函数
                self.operation = CustomOperation(name, Functions.custom_function[name])
                self.parse_children(argv)
            elif Functions.is_defined(name.lower()):
                # 在预置函数中有，归为自定义函数
                func = Functions.get_defined(name.lower())
                self.operation = CustomOperation(name, func)
                self.parse_children(argv)
            else:
                raise ParseError(f"unknown function: {name}")
        return self

    @staticmethod
    def _get_variable_value(field, params):
        """
        计算时获取变量的值
        :param field: 变量名
        :param params: calc时传入的params
        :return: float
        """
        if field.lower() in defined_constant:
            val = defined_constant[field.lower()]
            return val
        # 传入的是dict
        if isinstance(params, dict):
            if field not in params:
                # 未传入该变量的值，去预定义常量中查找
                if field.lower() in defined_constant:
                    val = defined_constant[field.lower()]
                    return val
                raise CalcError('not found value of "%s"' % field)
            val = params[field]
            if isinstance(val, str):
                val = float(val)
            return val
        # 传入的是function
        if callable(params):
            return params(field)
        raise CalcError('value error')

    def get_value(self, params):
        """
        calc时，获取该节点的计算结果
        :param params: calc传入的params
        :return: float
        """
        # 此节点为常量，直接取值
        if self.value is not None:
            return float(self.value)

        if self.operation is None:
            raise CalcError('string not parse yet')

        # 此节点为变量，直接获取该变量的值
        if hasattr(self.operation, 'variable'):
            field = self.operation.variable
            return self._get_variable_value(field, params)

        # 获取该节点所有child的值，作为参数带入operation求值
        argv = []
        for i in range(0, len(self.children)):
            val = self.children[i].get_value(params)
            argv.append(val)
        try:
            value = self.operation.operate(argv)
        except CalcError:
            return None
        return value



from abc import abstractmethod

from lexp.exception import CalcError


class Operation:
    # operations
    # 以下为定义的各种运算方法
    # 各类算数、逻辑运算符，以及自定义函数，都被定义为operation

    # 运算名称
    name = ''
    # 所需最少参数个数
    argc = 0

    def __init__(self):
        pass

    def check(self, argv):
        """
        在每个operation计算之前检查参数个数
        :param argv: 参数列表
        :return: None
        """
        if len(argv) < self.argc:
            raise CalcError('function %s: %d arguments need' % (self.name, self.argc), self, argv)

    @abstractmethod
    def operate(self, argv):
        """
        带入所需参数，使用该operation得到结果
        :param argv: 参数列表
        :return: value(float)
        """
        raise NotImplementedError



from lexp.exception import CalcError
from lexp.operation import Operation


class Bracket(Operation):
    # 括号
    def __init__(self):
        super().__init__()
        self.argc = 1

    def operate(self, argv):
        return float(argv[0])


class Variable(Operation):
    # 变量
    def __init__(self, variable):
        super().__init__()
        self.variable = variable

    def operate(self, val):
        return float(val)


class CustomOperation(Operation):
    # 自定义函数
    def __init__(self, name, func):
        super().__init__()
        self.name = name
        self.func = func
        if self.func:
            self.argc = func.__code__.co_argcount
        else:
            self.argc = None

    def operate(self, argv):
        if self.argc is not None:
            super().check(argv)
        if callable(self.func) is False:
            raise CalcError('%s not callable' % self.name)
        result = self.func(*argv)
        if isinstance(result, str):
            result = float(result)
        return result



from lexp.operation import Operation

# 一元运算符


class Not(Operation):
    # 非运算
    name = '!'
    argc = 1

    def operate(self, argv):
        super().check(argv)
        return 1 if float(argv[0]) == 0 else 0


class Negative(Operation):
    # 负号运算
    name = '-'
    argc = 1

    def operate(self, argv):
        super().check(argv)
        return 0 - float(argv[0])



from lexp.operation import Operation

# 二元运算符，只返回1/0


class Bool(Operation):
    argc = 2

    def __init__(self, func):
        # func为成立条件，返回True/False
        super().__init__()
        self.func = func

    def operate(self, argv):
        super().check(argv)
        val0 = None if argv[0] is None else float(argv[0])
        val1 = None if argv[1] is None else float(argv[1])
        return 1 if self.func(val0, val1) else 0


class And(Bool):
    name = '&&'

    def __init__(self):
        super().__init__(lambda x, y: x != 0 and y != 0)


class Or(Bool):
    name = '||'

    def __init__(self):
        super().__init__(lambda x, y: x != 0 or y != 0)


class Eq(Bool):
    name = '=='

    def __init__(self):
        super().__init__(lambda x, y: x == y)


class Neq(Bool):
    name = '!='

    def __init__(self):
        super().__init__(lambda x, y: x != y)


class Gt(Bool):
    name = '>'

    def __init__(self):
        super().__init__(lambda x, y: x > y)


class Lt(Bool):
    name = '<'

    def __init__(self):
        super().__init__(lambda x, y: x < y)


class Ge(Bool):
    name = '>='

    def __init__(self):
        super().__init__(lambda x, y: x >= y)


class Le(Bool):
    name = '<='

    def __init__(self):
        super().__init__(lambda x, y: x <= y)



# -*- coding: utf-8 -*-
import typing
from copy import deepcopy

from lexp.exception import CalcError
from lexp.expnode import _Node
from lexp.funcs import Functions


class Lexp:
    """
    实例化后可解析表达式，进行计算，如
    exp = Lexp()
    exp.load('expression string')
    result = exp.calc({'x': x value, 'y': y value})
    """
    def __init__(self, string=None):
        self.string = None
        self._expression = None
        if isinstance(string, str):
            self.load(string)

    def load(self, string):
        """
        解析表达式
        :param string: 表达式字符串
        :return: 成功与否
        """
        if not string:
            self._expression = None
            return False
        self.string = string
        self._expression = _Node(string)
        self._expression.parse()
        # self._print_expresstion(self._expression)
        return False if self._expression is None else True

    def calc(self, callback=None, params=None, **kwargs):
        """
        用已经解析的表达式，带入变量值或可获得值的函数，得到计算结果
        :param callback: callable, 带入变量名，返回变量值
        :param params: dict, 带入变量名，获得变量值
        :param kwargs: 变量值
        :return: float, 计算结果
        """
        if self._expression is None:
            raise CalcError("no expression be loaded")

        if callback is not None:
            return self._expression.get_value(callback)

        if params is None:
            p = dict()
        else:
            p = deepcopy(params)
        p.update(kwargs)
        return self._expression.get_value(p)

    def get_functions(self, only_undefined=False):
        """
        获得当前表达式所调用的函数
        :return: list
        """
        if self._expression is None:
            # raise CalcError("no expression be loaded")
            return list()
        functions = self._expression.get_functions()
        if only_undefined:
            functions = list(filter(lambda x: x.get("type") == "undefined", functions))
        return functions

    def get_variable_list(self):
        """
        获得当前表达式中出现的所有变量
        :return: list
        """
        if self._expression is None:
            # raise CalcError("no expression be loaded")
            return list()
        return self._expression.getVariables()

    def is_available(self):
        """
        当前是否已成功读取表达式，可进行计算
        :return: True/False
        """
        if self._expression is None:
            return False
        return True

    def __str__(self):
        return "" if self.string is None else self.string

    @classmethod
    def register(cls, name, func):
        """
        添加自定义函数，为静态方法
        :param name: 表达式中的函数名
        :param func: 函数
        :return: None
        """
        if callable(func) is False:
            raise Exception(f"{name} not callable")
        Functions.custom_function[name] = func

    # @classmethod
    # def _print_expresstion(cls, node, deep=0):
    #     """
    #     print解析后的层级关系，调试使用
    #     :param node: 需print的节点
    #     :param deep: 当前node的缩进数
    #     :return: None
    #     """
    #     if node is None:
    #         return
    #     d = deep + 1
    #     s = ''
    #     for i in range(0, deep):
    #         s = s + '\t'
    #     if node.value is not None:
    #         v = node.value
    #     else:
    #         t = str(type(node.operation))
    #         find = 'Lexp.Node.'
    #         p0 = t.find(find)
    #         if p0 != -1:
    #             t = t[p0 + len(find):-2]
    #         v = "%s %s" % (node.string, t)
    #     s = s + v
    #     print(s)
    #     for i in range(0, len(node.children)):
    #         cls._print_expresstion(node.children[i], d)



from lexp.operation import Operation

# 三元运算符


class EitherOr(Operation):
    # a ? b : c
    name = '?:'
    argc = 3

    def operate(self, argv):
        super().check(argv)
        if float(argv[0]) == 0:
            return float(argv[2])
        else:
            return float(argv[1])



from setuptools import setup, find_packages


setup(
    name='lexp',
    version='0.1.0',
    description='calculate the value of the expression',
    author='Li Zhiyuan',
    author_email='nxdlzy@163.com',
    license='MIT',
    packages=find_packages()
)




class ParseError(Exception):
    def __init__(self, message):
        super().__init__(message)


class CalcError(Exception):
    def __init__(self, text, o=None, argv=None):
        self.info = {
            'name': o.name if o else '',
            'argc': o.argc if o else '',
            'argv': argv
        }
        super().__init__(text)
